------------------------------------------------------------------------
----                                                                ----
---- WF1160 IP Core                                                 ----
----                                                                ----
---- Description:                                                   ----
---- This model provides an embedded Universal Serial Bus host      ----
---- controller compatible to the Philips ISP1160.                  ----
----                                                                ----
---- This entity is the USB host controller.                        ----
---- It implements the protocol layer of the USB 2.0 specification  ----
---- for low speed and full speed devices and attaches to the       ----
---- ISP160 fifo structure and its registers.                       ----
----                                                                ----
---- This host controller works with high speed signaling. The root ----
---- hub provides the signal conversion to lo speed when required.  ----
----                                                                ----
---- Lo speed device  Control sign   Hi speed device                ----
----                                                                ----
---- DP  DM           CTRL           DP  DM                         ----
----  0   0           SE0             0   0  -- Single ended zero.  ----
----  0   1            J              1   0                         ----
----  1   0            K              0   1                         ----
----  1   1           SE1             1   1  -- Single ended one.   ----
----                                                                ----
------------------------------------------------------------------------
----                                                                ----
----                                                                ----
----                                                                ----
---- Author(s):                                                     ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de     ----
----                                                                ----
------------------------------------------------------------------------
----                                                                ----
---- Copyright © 2020... Wolfgang Foerster - Inventronik GmbH.      ----
----                                                                ----
---- This source file may be used and distributed without           ----
---- restriction provided that this copyright statement is not      ----
---- removed from the file and that any derivative work contains    ----
---- the original copyright notice and the associated disclaimer.   ----
----                                                                ----
---- This source file is free software; you can redistribute it     ----
---- and/or modify it under the terms of the GNU Lesser General     ----
---- Public License as published by the Free Software Foundation;   ----
---- either version 2.1 of the License, or (at your option) any     ----
---- later version.                                                 ----
----                                                                ----
---- This source is distributed in the hope that it will be         ----
---- useful, but WITHOUT ANY WARRANTY; without even the implied     ----
---- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR        ----
---- PURPOSE. See the GNU Lesser General Public License for more    ----
---- details.                                                       ----
----                                                                ----
---- You should have received a copy of the GNU Lesser General      ----
---- Public License along with this source; if not, download it     ----
---- from http://www.gnu.org/licenses/lgpl.html                     ----
----                                                                ----
------------------------------------------------------------------------
--
-- Revision History
--
-- Revision 2K20B  20201224 WF
--   Initial Release.
--

library work;
use work.wf1160_pkg.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

entity HOST_CONTROLLER is
    port (
        CLK_48MHz           : in std_logic;
        RESET               : in std_logic;

        CSn                 : in std_logic;

        -- USB data:
        DP_IN               : in std_logic; -- Data line.
        DM_IN               : in std_logic; -- Data line.
        DP_OUT              : out std_logic; -- Data line.
        DM_OUT              : out std_logic; -- Data line.
        DPM_EN              : out std_logic; -- Output enable.
        SPEED_Tx            : out std_logic; -- '1' for high speed data.

        -- Port Status and control:
        RWE                 : in std_logic; -- Remote wakeup enable.
        HC_RESUME           : in std_logic; -- DRWE controlled wakeup enable.
        WAKEUP              : in std_logic; -- Wakeup from suspend.
        SUSPEND             : out std_logic; -- Suspend status.

        -- Buffer
        ATL_BUFF_LEN        : in std_logic_vector(11 downto 0);
        ITL_BUFF_LEN        : in std_logic_vector(11 downto 0);
        BUFFER_IN           : in std_logic_vector(15 downto 0);
        BUFFER_OUT          : out std_logic_vector(15 downto 0);

        -- Buffer status:
        ATL_BUFF_FULL       : in std_logic;
        ITL0_BUFF_FULL      : in std_logic;
        ITL1_BUFF_FULL      : in std_logic;
        ATL_BUFF_DONE       : out std_logic;
        ITL0_BUFF_DONE      : out std_logic;
        ITL1_BUFF_DONE      : out std_logic;

        -- FIFO_status: 
        ATL_FIFO_FULL       : in std_logic;
        ITL0_FIFO_FULL      : in std_logic;
        ITL1_FIFO_FULL      : in std_logic;
        ATL_FIFO_EMPTY      : in std_logic;
        ITL0_FIFO_EMPTY     : in std_logic;
        ITL1_FIFO_EMPTY     : in std_logic;

        -- Buffer control:
        ATL_RD              : buffer std_logic;
        ATL_WR              : buffer std_logic;
        ITL_RD              : buffer std_logic;
        ITL_WR              : buffer std_logic;
        HC_ITL1             : buffer std_logic; -- ITL1 in use by the host controller when '1'.
        HC_ADR              : out std_logic_vector(11 downto 0);

        -- Controls:
        HCR                 : in std_logic; -- Host controller reset.
        SOFTWARE_RESET      : in std_logic;
        HCFS_IN             : in std_logic_vector(1 downto 0); -- Host controller functional state.
        HCFS_OUT            : out std_logic_vector(1 downto 0); -- Host controller functional state.
        SOC                 : buffer std_logic_vector(1 downto 0); -- Scheduling overrun count.
        PSSC_1              : in std_logic; -- Port suspend status change.
        PSSC_2              : in std_logic; -- Port suspend status change.

        -- Not used controls:
        CLKNS               : in std_logic; -- Suspend CLK not stop.
        AOCEN               : in std_logic; -- Analog OC enable.

        FSMPS               : in std_logic_vector(14 downto 0); -- FS largest data packet.
        FR                  : in std_logic_vector(13 downto 0); -- Frame interval.
        FR_DEC              : out std_logic; -- Decrement frame remaining.
        LST                 : in std_logic_vector(10 downto 0); -- LSThreshold.

        HC_SUSPND           : out std_logic; -- Host controller suspended.
        ATL_INT             : out std_logic; -- ATL data to be read.

        UE                  : out std_logic; -- Unrecoverable error.
        RD                  : out std_logic; -- Resume detected.
        SF                  : out std_logic; -- Start of frame.
        SO                  : out std_logic; -- Scheduling overrun.

        OPBERR              : out std_logic -- Operational bus error.
    );
end entity HOST_CONTROLLER;

architecture BEHAVIOUR of HOST_CONTROLLER is
type PTD_HEADER_TYPE is array(0 to 7) of std_logic_vector(7 downto 0);
signal PTD_HEADER           : PTD_HEADER_TYPE;
signal ACTUAL_BYTES         : std_logic_vector(9 downto 0);
signal MAX_PACKET_SIZE      : std_logic_vector(9 downto 0);
signal TOTAL_BYTES          : std_logic_vector(9 downto 0);

alias ACTIVE                : std_logic is PTD_HEADER(1)(3);
alias TOGGLE                : std_logic is PTD_HEADER(1)(2);
alias ENDPOINT_NUMBER       : std_logic_vector(3 downto 0) is PTD_HEADER(3)(7 downto 4);
alias LAST                  : std_logic is PTD_HEADER(3)(3);
alias SPEED                 : std_logic is PTD_HEADER(3)(2);
alias B5_5                  : std_logic is PTD_HEADER(5)(5);
alias DIRECTION_PID         : std_logic_vector(1 downto 0) is PTD_HEADER(5)(3 downto 2);
alias FORMAT                : std_logic is PTD_HEADER(6)(7); -- Isochronous = '1'.
alias FUNCTION_ADDRESS      : std_logic_vector(6 downto 0) is PTD_HEADER(6)(6 downto 0);

type HC_STATES is (USB_RESET, USB_RESUME, USB_SUSPEND, USB_OPERATIONAL, READ_ITL_PTD, READ_ATL_PTD, TRANSMIT_ITL_DATA, 
                   TRANSMIT_ATL_DATA, RECEIVE_ITL_DATA, RECEIVE_ATL_DATA, TOKEN_SOF, PREAMBLE, WRITE_ITL_PTD, WRITE_ATL_PTD);
type SIE_STATES is(IDLE, HANDSHAKE_Rx, HANDSHAKE_Tx, DATA_EOP, DATA_PID_Rx, DATA_PID_Tx, DATA_Tx, DATA_Rx, SYNC, SYNC_DATA, TOKEN_EOP, TOKEN_PID, TOKEN_Tx, WAIT_HUBSETUP);
signal HC_STATE             : HC_STATES;
signal HC_NSTATE            : HC_STATES; -- Next state.
signal SIE_STATE            : SIE_STATES;
signal SIE_NSTATE           : SIE_STATES;
signal ATL_PENDING          : std_logic_vector(6 downto 0);
signal BUBBLE_LOA           : std_logic;
signal CS_I                 : std_logic;
signal DATA_RDY             : std_logic;
signal SIE_RDY              : std_logic;
signal CRC_ERROR            : std_logic;
signal SOF                  : std_logic;
signal WAKEUP_I             : std_logic;
signal PID                  : std_logic_vector(3 downto 0);
signal PTD_CNT              : std_logic_vector(2 downto 0);
signal Tx_STRB              : std_logic;
signal STUFF_ERR            : std_logic;
signal Tx_RDY               : std_logic;
signal Rx_RDY               : std_logic;
signal Rx_STRB              : std_logic;
signal Rx_HOLD_REG          : std_logic_vector(15 downto 0);
signal FRAME_NUMBER         : std_logic_vector(10 downto 0);
signal RD_I                 : std_logic;
signal SETUP_RDY            : std_logic;
signal TIMEOUT              : std_logic;
signal WRONG_EOP            : std_logic;
begin
    P_SYNC: process
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        WAKEUP_I <= WAKEUP;
        CS_I <= not CSn;
    end process P_SYNC;

    SOF_FRAME: process
    -- This process is responsible for the SOF generation and
    -- the frame number handling.
    variable CNT_1ms    : integer range 0 to 65535;
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        
        SOF <= '0'; -- Strobe.
        
        if HC_STATE = USB_RESET or HC_STATE = USB_SUSPEND or HC_STATE = USB_RESUME then
            CNT_1ms := 0;
            FRAME_NUMBER <= "00000000000";
        elsif CNT_1ms < 47999 then
            CNT_1ms := CNT_1ms + 1;
        else
            CNT_1ms := 0;
            FRAME_NUMBER <= FRAME_NUMBER + '1';
            SOF <= '1';
        end if;
    end process SOF_FRAME;

    SF <= SOF;
    SIE_RDY <= '1' when SIE_STATE /= IDLE and SIE_NSTATE = IDLE else '0';

    HC_STATEREG: process
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if RESET  = '1' or SOFTWARE_RESET = '1' then
            HC_STATE <= USB_RESET;
        else
            HC_STATE <= HC_NSTATE;
        end if;
    end process HC_STATEREG;

    HC_STATEDEC: process(ATL_PENDING, CS_I, DIRECTION_PID, FORMAT, FR, FSMPS, HC_STATE, HCFS_IN, HCR, HC_RESUME, ITL0_BUFF_FULL, LST, SIE_RDY, SOF, ITL1_BUFF_FULL, ATL_BUFF_FULL, PSSC_1, PSSC_2, PTD_CNT, RD_I, RWE, WAKEUP_I)
    begin
        case HC_STATE is
            when USB_RESET =>
                if HCFS_IN = "10" then
                    HC_NSTATE <= USB_OPERATIONAL;
                else
                    HC_NSTATE <= USB_RESET;
                end if;
            when USB_SUSPEND =>
                if HCFS_IN = "00" then
                    HC_NSTATE <= USB_RESET;
                elsif HCFS_IN = "01" then
                    HC_NSTATE <= USB_RESUME;
                elsif HCFS_IN = "10" then
                    HC_NSTATE <= USB_OPERATIONAL;
                elsif WAKEUP_I = '1' then
                    HC_NSTATE <= USB_RESUME;
                elsif CS_I = '1' then
                    HC_NSTATE <= USB_RESUME;
                elsif RWE = '1' and RD_I = '1' then
                    HC_NSTATE <= USB_RESUME;
                elsif HC_RESUME = '1' then
                    HC_NSTATE <= USB_RESUME;
                else
                    HC_NSTATE <= USB_SUSPEND;
                end if;
            when USB_RESUME =>
                if HCFS_IN = "00" then
                    HC_NSTATE <= USB_RESET;
                elsif HCFS_IN = "10" then
                    HC_NSTATE <= USB_OPERATIONAL;
                elsif WAKEUP_I = '0' then
                    HC_NSTATE <= USB_SUSPEND;
                else
                    HC_NSTATE <= USB_RESUME;
                end if;
            when USB_OPERATIONAL =>
                if HCFS_IN = "00" then
                    HC_NSTATE <= USB_RESET;
                elsif HCFS_IN = "11" then
                    HC_NSTATE <= USB_SUSPEND;
                elsif HCR = '1' then
                    HC_NSTATE <= USB_SUSPEND;
                elsif SOF = '1' then
                    HC_NSTATE <= TOKEN_SOF;
                else
                    HC_NSTATE <= USB_OPERATIONAL;
                end if;
            when TOKEN_SOF =>
                if SIE_RDY = '1' and ITL0_BUFF_FULL = '1' then
                    HC_NSTATE <= READ_ITL_PTD;
                elsif SIE_RDY = '1' and ITL1_BUFF_FULL = '1' then
                    HC_NSTATE <= READ_ITL_PTD;
                elsif SIE_RDY = '1' and ATL_BUFF_FULL = '1' then
                    HC_NSTATE <= READ_ATL_PTD;
                elsif SIE_RDY = '1' then -- Nothing to do.
                    HC_NSTATE <= USB_OPERATIONAL;
                else
                    HC_NSTATE <= TOKEN_SOF;
                end if;
            when READ_ITL_PTD =>
                if PTD_CNT = "100" and ACTIVE = '0' and LAST = '1' and ATL_BUFF_FULL = '1' then
                    HC_NSTATE <= READ_ATL_PTD; -- Go on with ATL buffer.
                elsif PTD_CNT = "100" and ACTIVE = '0' and LAST = '1' then
                    HC_NSTATE <= USB_OPERATIONAL; -- Nothing to do.
                elsif PTD_CNT = "100" and ACTIVE = '0' then
                    HC_NSTATE <= READ_ITL_PTD; -- Next header.
                -- There are no Iisochronous low speed transfers in the USB
                -- specification.
                elsif PTD_CNT = "100" and ACTIVE = '1' then
                    case DIRECTION_PID is
                        -- when "00" => -- No control tokens during ISO.
                        when "01" => HC_NSTATE <= TRANSMIT_ITL_DATA;
                        when "10" => HC_NSTATE <= RECEIVE_ITL_DATA;
                        when others => HC_NSTATE <= USB_OPERATIONAL; -- Reserved, never true or error.
                    end case;
                else
                    HC_NSTATE <= READ_ITL_PTD;
                end if;
            when READ_ATL_PTD =>
                if PTD_CNT = "100" and ACTIVE = '0' and LAST = '1' and ATL_PENDING = "0000000" then
                    HC_NSTATE <= USB_OPERATIONAL; -- Nothing to do, go idle.
                elsif PTD_CNT = "100" and FR < FSMPS then
                    HC_NSTATE <= USB_OPERATIONAL; -- No time left, go idle.
                elsif PTD_CNT = "100" and ACTIVE = '0' then
                    HC_NSTATE <= READ_ATL_PTD; -- Next header.
                elsif PTD_CNT = "100" and ACTIVE = '1' and SPEED = '0' and FR < LST then -- Low speed device.
                    HC_NSTATE <= READ_ATL_PTD; -- Reject the access due to low speed threshold violation.
                elsif PTD_CNT = "100" and ACTIVE = '1' and SPEED = '0' then -- Low speed device.
                    HC_NSTATE <= PREAMBLE;
                elsif PTD_CNT = "100" and ACTIVE = '1' then
                    case DIRECTION_PID is
                        when "00" => HC_NSTATE <= TRANSMIT_ATL_DATA;
                        when "01" => HC_NSTATE <= TRANSMIT_ATL_DATA;
                        when "10" => HC_NSTATE <= RECEIVE_ATL_DATA;
                        when others => HC_NSTATE <= USB_OPERATIONAL; -- Reserved, never true or error.
                    end case;
                else
                    HC_NSTATE <= READ_ATL_PTD;
                end if;
            when PREAMBLE => -- Low speed preamble.
                if SIE_RDY = '1' then
                    case DIRECTION_PID is
                        when "00" => HC_NSTATE <= TRANSMIT_ATL_DATA;
                        when "01" => HC_NSTATE <= TRANSMIT_ATL_DATA;
                        when "10" => HC_NSTATE <= RECEIVE_ATL_DATA;
                        when others => HC_NSTATE <= USB_OPERATIONAL; -- Reserved, never true or error.;
                    end case;
                else
                    HC_NSTATE <= PREAMBLE;
                end if;
            when TRANSMIT_ITL_DATA =>
                if SIE_RDY = '1' then
                    HC_NSTATE <= WRITE_ITL_PTD;
                else
                    HC_NSTATE <= TRANSMIT_ITL_DATA;
                end if;
            when TRANSMIT_ATL_DATA =>
                if SIE_RDY = '1' then
                    HC_NSTATE <= WRITE_ATL_PTD;
                else
                    HC_NSTATE <= TRANSMIT_ATL_DATA;
                end if;
            when RECEIVE_ITL_DATA =>
                if SIE_RDY = '1' then
                    HC_NSTATE <= WRITE_ITL_PTD;
                else
                    HC_NSTATE <= RECEIVE_ITL_DATA;
                end if;
            when RECEIVE_ATL_DATA =>
                if SIE_RDY = '1' then
                    HC_NSTATE <= WRITE_ATL_PTD;
                else
                    HC_NSTATE <= RECEIVE_ATL_DATA;
                end if;
            when WRITE_ITL_PTD =>
                if PTD_CNT = "100" and LAST = '1' and ATL_BUFF_FULL = '1' then
                    HC_NSTATE <= READ_ATL_PTD;
                elsif PTD_CNT = "100" and LAST = '1' then
                    HC_NSTATE <= USB_OPERATIONAL; -- Nothing to do.
                elsif PTD_CNT = "100" then
                    HC_NSTATE <= READ_ITL_PTD;
                else
                    HC_NSTATE <= WRITE_ITL_PTD;
                end if;
            when WRITE_ATL_PTD =>
                if PTD_CNT = "100" and B5_5 = '1' then
                    HC_NSTATE <= USB_OPERATIONAL; -- Only this transaction in 1ms frame.
                elsif PTD_CNT = "100" and LAST = '1' and ATL_PENDING = "0000000" then
                    HC_NSTATE <= USB_OPERATIONAL; -- Nothing to do, go idle.
                elsif PTD_CNT = "100" then
                    HC_NSTATE <= READ_ATL_PTD;
                else
                    HC_NSTATE <= WRITE_ATL_PTD;
                end if;
        end case;
    end process HC_STATEDEC;

     -- Early indication for HCFS to keep the host controller
     -- state machine and the controll register synchronous.
    with HC_NSTATE select
        HCFS_OUT <= "00" when USB_RESET,
                    "01" when USB_RESUME,
                    "11" when USB_SUSPEND,
                    "10" when others;

    HC_SUSPND <= '1' when HC_STATE = USB_SUSPEND else '0';

    BUFFER_CONTROL: process
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if HC_STATE = USB_RESET or HC_STATE = USB_SUSPEND or HC_STATE = USB_RESUME then
            HC_ITL1 <= '0'; -- Host controller accesses ITL0.
        elsif ITL0_BUFF_FULL  = '1' and ITL1_BUFF_FULL  = '1' then
            HC_ITL1 <= '0'; -- Host controller accesses ITL0.
        elsif SOF = '1' then
            HC_ITL1 <= not HC_ITL1;
        end if;

        if SOF = '1' then
            ITL1_BUFF_DONE <= '0';
            ITL0_BUFF_DONE <= '0';
            ATL_BUFF_DONE <= '0';
        elsif HC_STATE = WRITE_ITL_PTD and HC_NSTATE = USB_OPERATIONAL and HC_ITL1 = '1' then
            ITL1_BUFF_DONE <= '1';
        elsif HC_STATE = WRITE_ITL_PTD and HC_NSTATE = USB_OPERATIONAL then
            ITL0_BUFF_DONE <= '1';
        elsif HC_STATE = WRITE_ATL_PTD and HC_NSTATE = USB_OPERATIONAL then
            ATL_BUFF_DONE <= '1';
        end if;        
    end process BUFFER_CONTROL;

    FIFO_CONTROL: process
    -- This logic controls the FIFO read and write signals for both, 
    -- ITL and ATL. It also handles the PTD header related data such 
    -- as the header data and the completion codes.
    variable ADR_VAR        : integer range 0 to 7;
    variable CLK_CNT        : std_logic_vector(2 downto 0);
    variable HC_ADRVAR      : std_logic_vector(11 downto 0);
    variable PTD_ENTRY      : std_logic_vector(11 downto 0);
    variable SCND_SCAN      : boolean;
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;

        -- PTD header counter:
        if HC_STATE /= READ_ITL_PTD and HC_NSTATE = READ_ITL_PTD then
            PTD_CNT <= "000";
            CLK_CNT := "000";
        elsif HC_STATE /= READ_ATL_PTD and HC_NSTATE = READ_ATL_PTD then
            PTD_CNT <= "000";
            CLK_CNT := "000";
        elsif HC_STATE /= WRITE_ITL_PTD and HC_NSTATE = WRITE_ITL_PTD then
            PTD_CNT <= "000";
            CLK_CNT := "000";
        elsif HC_STATE /= WRITE_ATL_PTD and HC_NSTATE = WRITE_ATL_PTD then
            PTD_CNT <= "000";
            CLK_CNT := "000";
        elsif (HC_STATE = READ_ATL_PTD or HC_STATE = READ_ITL_PTD or HC_STATE = WRITE_ATL_PTD or HC_STATE = WRITE_ITL_PTD) and CLK_CNT = "101" then
            PTD_CNT <= PTD_CNT + '1';
            CLK_CNT := "000";
        else -- We need 6 clock cycles to access a fifo entry.
             -- For ease of implementation we use 4 clock cycles.
            CLK_CNT := CLK_CNT + '1';
            --
            if PTD_CNT = "100" then
                PTD_CNT <= "000"; -- Rollover for next PTD header access.
            end if;
        end if;

        ATL_RD <= '0'; -- Strobe.
        ATL_WR <= '0'; -- Strobe.
        ITL_RD <= '0'; -- Strobe.
        ITL_WR <= '0'; -- Strobe.

        -- Read/write controls:
        if HC_STATE /= READ_ITL_PTD and HC_NSTATE = READ_ITL_PTD then
            ITL_RD <= '1';
        elsif HC_STATE /= READ_ATL_PTD and HC_NSTATE = READ_ATL_PTD then
            ATL_RD <= '1';
        elsif HC_STATE /= TRANSMIT_ITL_DATA and HC_NSTATE = TRANSMIT_ITL_DATA then
            ITL_RD <= '1';
        elsif HC_STATE /= TRANSMIT_ATL_DATA and HC_NSTATE = TRANSMIT_ATL_DATA then
            ATL_RD <= '1';
        elsif HC_STATE = READ_ATL_PTD and HC_NSTATE = READ_ATL_PTD and CLK_CNT = "101" then
            ATL_RD <= '1';
        elsif HC_STATE = READ_ITL_PTD and HC_NSTATE = READ_ITL_PTD and CLK_CNT = "101" then
            ITL_RD <= '1';
        elsif HC_STATE = WRITE_ATL_PTD and HC_NSTATE = WRITE_ATL_PTD and CLK_CNT = "101" then
            ATL_WR <= '1';
        elsif HC_STATE = WRITE_ITL_PTD and HC_NSTATE = WRITE_ITL_PTD and CLK_CNT = "101" then
            ITL_WR <= '1';
        elsif HC_STATE = TRANSMIT_ITL_DATA and SIE_STATE = DATA_Tx and DATA_RDY = '1' then
            ITL_RD <= '1';
        elsif HC_STATE = TRANSMIT_ATL_DATA and SIE_STATE = DATA_Tx and DATA_RDY = '1' then
            ATL_RD <= '1';
        elsif HC_STATE = RECEIVE_ITL_DATA and SIE_STATE = DATA_Rx and DATA_RDY = '1' then
            ITL_WR <= '1';
        elsif HC_STATE = RECEIVE_ATL_DATA and SIE_STATE = DATA_Rx and DATA_RDY = '1' then
            ATL_WR <= '1';
        end if;

        -- Address logic:
        if HC_STATE = USB_RESET then
            HC_ADRVAR := x"000";
        elsif HC_STATE /= READ_ITL_PTD and HC_NSTATE = READ_ITL_PTD then
            HC_ADRVAR := x"000"; -- Initialize.
        elsif HC_STATE /= READ_ATL_PTD and HC_NSTATE = READ_ATL_PTD then
            HC_ADRVAR := x"000"; -- Initialize.
        elsif HC_STATE = READ_ATL_PTD and HC_NSTATE = READ_ATL_PTD and PTD_CNT = "100" and LAST = '1' then
            HC_ADRVAR := x"000"; -- Reset to tail.
        elsif HC_STATE = READ_ATL_PTD and HC_NSTATE = READ_ATL_PTD and PTD_CNT = "100" then
            HC_ADRVAR := HC_ADRVAR + TOTAL_BYTES + "10"; -- Go to next header.
        elsif HC_STATE = READ_ATL_PTD and PTD_CNT = "100" then
            HC_ADRVAR := HC_ADRVAR + ACTUAL_BYTES; -- Go to payload address (valid for split or non split packages).
        elsif HC_STATE = READ_ATL_PTD and CLK_CNT = "101" then
            HC_ADRVAR := HC_ADRVAR + "10"; -- Word wide.
        elsif HC_STATE = READ_ITL_PTD and HC_NSTATE = READ_ITL_PTD and PTD_CNT = "100" then
            HC_ADRVAR := HC_ADRVAR + MAX_PACKET_SIZE + "10"; -- Go to next header.
        elsif HC_STATE = READ_ITL_PTD and CLK_CNT = "101" then
            HC_ADRVAR := HC_ADRVAR + "10"; -- Word wide.
        elsif HC_STATE = RECEIVE_ATL_DATA then
            if Rx_RDY = '1' and HC_ADRVAR(1 downto 0) = "10" then -- After incrementing we are on a non double word boundary.
                HC_ADRVAR := HC_ADRVAR + "10"; -- Switch to double word boundary.
            elsif DATA_RDY = '1' then -- After incrementing we are on a double word boundary.
                HC_ADRVAR := HC_ADRVAR + "10"; -- Word wide.
            end if;
        elsif HC_STATE = TRANSMIT_ATL_DATA then
            if Tx_RDY = '1' and HC_ADRVAR(1 downto 0) = "10" then -- After incrementing we are on a non double word boundary.
                HC_ADRVAR := HC_ADRVAR + "10"; -- Switch to double word boundary.
            elsif DATA_RDY = '1' then -- After incrementing we are on a double word boundary.
                HC_ADRVAR := HC_ADRVAR + "10"; -- Word wide.
            end if;
        elsif HC_STATE = RECEIVE_ITL_DATA then
            if Rx_RDY = '1' and HC_ADRVAR(1 downto 0) = "10" then -- After incrementing we are on a non double word boundary.
                HC_ADRVAR := HC_ADRVAR + "10"; -- Switch to double word boundary.
            elsif DATA_RDY = '1' then -- After incrementing we are on a double word boundary.
                HC_ADRVAR := HC_ADRVAR + "10"; -- Word wide.
            end if;
        elsif HC_STATE = TRANSMIT_ITL_DATA then
            if Tx_RDY = '1' and HC_ADRVAR(1 downto 0) = "10" then -- After incrementing we are on a non double word boundary.
                HC_ADRVAR := HC_ADRVAR + "10"; -- Switch to double word boundary.
            elsif DATA_RDY = '1' then -- After incrementing we are on a double word boundary.
                HC_ADRVAR := HC_ADRVAR + "10"; -- Word wide.
            end if;
        end if;

        if HC_STATE /= READ_ATL_PTD and HC_NSTATE = READ_ATL_PTD then
            PTD_ENTRY := HC_ADRVAR; -- Store the entry point.
        elsif HC_STATE = READ_ATL_PTD and HC_NSTATE = READ_ATL_PTD and PTD_CNT = "100" then
            PTD_ENTRY := HC_ADRVAR; -- Store the entry point of the next header.
        elsif HC_STATE = WRITE_ATL_PTD and CLK_CNT = "101" then
            PTD_ENTRY := PTD_ENTRY + "10"; -- Word wide.
        elsif HC_STATE /= READ_ITL_PTD and HC_NSTATE = READ_ITL_PTD then
            PTD_ENTRY := HC_ADRVAR; -- Store the entry point.
        elsif HC_STATE = READ_ATL_PTD and HC_NSTATE = READ_ITL_PTD and PTD_CNT = "100" then
            PTD_ENTRY := HC_ADRVAR; -- Store the entry point of the next header.
        elsif HC_STATE = WRITE_ITL_PTD and CLK_CNT = "101" then
            PTD_ENTRY := PTD_ENTRY + "10"; -- Word wide.
        end if;

        case HC_STATE is
            when WRITE_ATL_PTD | WRITE_ITL_PTD => HC_ADR <= PTD_ENTRY;
            when others => HC_ADR <= HC_ADRVAR;
        end case;

        -- PTD header data:
        ADR_VAR := To_Integer(unsigned(PTD_CNT(1 downto 0)));

        if (HC_STATE = READ_ATL_PTD or HC_STATE = READ_ITL_PTD) and CLK_CNT = "101" then
            PTD_HEADER(2 * ADR_VAR) <= BUFFER_IN(15 downto 8);
            PTD_HEADER(2 * ADR_VAR + 1) <= BUFFER_IN(7 downto 0);
        end if;

        -- Completion code:
        if CRC_ERROR = '1' then
            PTD_HEADER(1)(7 downto 4) <= x"1"; -- CRC.
        elsif STUFF_ERR = '1' then
            PTD_HEADER(1)(7 downto 4) <= x"2"; -- Bit stuffing.
        elsif SIE_STATE = DATA_PID_Rx and Rx_RDY = '1' and FORMAT = '0' and Rx_HOLD_REG(3 downto 0) /= x"3" and TOGGLE = '0' then
            PTD_HEADER(1)(7 downto 4) <= x"3"; -- Data toggle mismatch.
        elsif SIE_STATE = DATA_PID_Rx and Rx_RDY = '1' and FORMAT = '0' and Rx_HOLD_REG(3 downto 0) /= x"B" and TOGGLE = '1' then
            PTD_HEADER(1)(7 downto 4) <= x"3"; -- Data toggle mismatch.
        elsif SIE_STATE = DATA_PID_Rx and Rx_RDY = '1' and Rx_HOLD_REG(3 downto 0) = x"E" then
            PTD_HEADER(1)(7 downto 4) <= x"4"; -- Stall.
        elsif TIMEOUT = '1' then
            PTD_HEADER(1)(7 downto 4) <= "0101"; -- Device not responding.
        elsif SIE_STATE = DATA_PID_Rx and Rx_RDY = '1' and Rx_HOLD_REG(7 downto 4) /= Rx_HOLD_REG(3 downto 0) then
            PTD_HEADER(1)(7 downto 4) <= "0110"; -- PID check failure.
        elsif SIE_STATE = HANDSHAKE_Rx and Rx_RDY = '1' and Rx_HOLD_REG(7 downto 4) /= Rx_HOLD_REG(3 downto 0) then
            PTD_HEADER(1)(7 downto 4) <= "0110"; -- PID check failure.
        elsif SIE_STATE = DATA_PID_Rx and Rx_RDY = '1' and Rx_HOLD_REG(3 downto 0) /= x"B" and Rx_HOLD_REG(3 downto 0) /= x"3" then -- No data PID.
            PTD_HEADER(1)(7 downto 4) <= "0111"; -- Unexpected PID.
        elsif SIE_STATE = HANDSHAKE_Rx and Rx_RDY = '1' and Rx_HOLD_REG(3 downto 0) /= x"E" and Rx_HOLD_REG(3 downto 0) /= x"A" and Rx_HOLD_REG(3 downto 0) /= x"2" then -- No handshake PID.
            PTD_HEADER(1)(7 downto 4) <= "0111"; -- Unexpected PID.
        elsif HC_STATE /= WRITE_ITL_PTD and HC_NSTATE = WRITE_ITL_PTD and ACTUAL_BYTES > MAX_PACKET_SIZE then
            PTD_HEADER(1)(7 downto 4) <= "1000"; -- Data overrun.
        elsif HC_STATE /= WRITE_ATL_PTD and HC_NSTATE = WRITE_ATL_PTD and ACTUAL_BYTES > TOTAL_BYTES then
            PTD_HEADER(1)(7 downto 4) <= "1000"; -- Data overrun.
        elsif HC_STATE /= WRITE_ITL_PTD and HC_NSTATE = WRITE_ITL_PTD and ACTUAL_BYTES < MAX_PACKET_SIZE and ACTUAL_BYTES < TOTAL_BYTES then -- The last package my be less than MAX_PACKET_SIZE.
            PTD_HEADER(1)(7 downto 4) <= "1001"; -- Data underrun.
        elsif HC_STATE /= WRITE_ATL_PTD and HC_NSTATE = WRITE_ATL_PTD and ACTUAL_BYTES = TOTAL_BYTES then
            PTD_HEADER(1)(7 downto 4) <= "0000"; -- All is fine.
        elsif HC_STATE = RECEIVE_ITL_DATA and ACTUAL_BYTES > ITL_BUFF_LEN then
            PTD_HEADER(1)(7 downto 4) <= "1000"; -- Data overrun.
        elsif HC_STATE = RECEIVE_ATL_DATA and ACTUAL_BYTES > ATL_BUFF_LEN then
            PTD_HEADER(1)(7 downto 4) <= "1000"; -- Data overrun.
        elsif HC_STATE = RECEIVE_ITL_DATA and HC_ITL1 = '0' and ITL0_FIFO_FULL = '1' and ITL_WR = '1' then
            PTD_HEADER(1)(7 downto 4) <= "0010"; -- Buffer overrun.
        elsif HC_STATE = RECEIVE_ITL_DATA and HC_ITL1 = '1' and ITL1_FIFO_FULL = '1' and ITL_WR = '1' then
            PTD_HEADER(1)(7 downto 4) <= "0010"; -- Buffer overrun.
        elsif HC_STATE = RECEIVE_ATL_DATA and ATL_FIFO_FULL = '1' and ATL_WR = '1' then
            PTD_HEADER(1)(7 downto 4) <= "0010"; -- Buffer overrun.
        elsif HC_STATE = TRANSMIT_ITL_DATA and HC_ITL1 = '0' and ITL0_FIFO_EMPTY = '1' and ITL_RD = '1' then
            PTD_HEADER(1)(7 downto 4) <= "0010"; -- Buffer underrun.
        elsif HC_STATE = TRANSMIT_ITL_DATA and HC_ITL1 = '1' and ITL1_FIFO_EMPTY = '1' and ITL_RD = '1' then
            PTD_HEADER(1)(7 downto 4) <= "0010"; -- Buffer underrun.
        elsif HC_STATE = TRANSMIT_ATL_DATA and ATL_FIFO_EMPTY = '1' and ATL_RD = '1' then
            PTD_HEADER(1)(7 downto 4) <= "0010"; -- Buffer underrun.
        end if;

        -- This is the pending logic. The ATL_PENDING counter counts the active lists during the
        -- first list iteration. If the list is done, the counter is decremented by one. If there 
        -- are pending lists after the last list, the host controller goes on with list processing.
        -- This logic is valid for ATL type lists.
        if SOF = '1' then
            ATL_PENDING <= "0000000";
            SCND_SCAN := false;
        elsif HC_STATE = READ_ATL_PTD and PTD_CNT = "100" and ACTIVE = '1' and SCND_SCAN = false then
            ATL_PENDING <= ATL_PENDING + '1'; -- Count the active lists during the first ATL list iteration.
        elsif HC_STATE /= WRITE_ATL_PTD and HC_NSTATE = WRITE_ATL_PTD and ACTUAL_BYTES = TOTAL_BYTES then
            ATL_PENDING <= ATL_PENDING - '1';
        elsif HC_STATE = WRITE_ATL_PTD and HC_NSTATE /= WRITE_ATL_PTD and LAST = '1' then
            SCND_SCAN := true;
        end if;

        -- Active bit:
        if HC_STATE /= WRITE_ITL_PTD and HC_NSTATE = WRITE_ITL_PTD and ACTUAL_BYTES = MAX_PACKET_SIZE then
            PTD_HEADER(1)(3) <= '0'; -- Done.
        elsif HC_STATE /= WRITE_ATL_PTD and HC_NSTATE = WRITE_ATL_PTD and ACTUAL_BYTES = TOTAL_BYTES then
            PTD_HEADER(1)(3) <= '0'; -- Done.
        end if;

        -- Toggle feature:
        if SIE_STATE = TOKEN_PID and PID = "1101" then
            PTD_HEADER(1)(2) <= '0'; -- Initialize during setup token.
        elsif SIE_STATE = HANDSHAKE_Rx and Rx_RDY = '1' and Rx_HOLD_REG = x"D2" then -- Transmitter receives Ack.
            PTD_HEADER(1)(2) <= not PTD_HEADER(1)(2); -- Toggle.
        elsif SIE_STATE = HANDSHAKE_Tx and Tx_RDY = '1' and PTD_HEADER(1)(7 downto 4) = x"0" then -- Valid data, correct PID.
            PTD_HEADER(1)(2) <= not PTD_HEADER(1)(2); -- Toggle.
        end if;
    end process FIFO_CONTROL;

    BUFFER_OUT <= PTD_HEADER(0) & PTD_HEADER(1) when (HC_STATE = WRITE_ATL_PTD or HC_STATE = WRITE_ITL_PTD) and PTD_CNT = "000" else
                  PTD_HEADER(2) & PTD_HEADER(3) when (HC_STATE = WRITE_ATL_PTD or HC_STATE = WRITE_ITL_PTD) and PTD_CNT = "001" else
                  PTD_HEADER(4) & PTD_HEADER(5) when (HC_STATE = WRITE_ATL_PTD or HC_STATE = WRITE_ITL_PTD) and PTD_CNT = "010" else
                  PTD_HEADER(6) & PTD_HEADER(7) when (HC_STATE = WRITE_ATL_PTD or HC_STATE = WRITE_ITL_PTD) and PTD_CNT = "011" else Rx_HOLD_REG;

    SUSPENDED: process
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if HC_STATE = USB_RESET or HC_STATE = USB_OPERATIONAL then
            SUSPEND <= '0';
        elsif HC_STATE = USB_SUSPEND then
            SUSPEND <= '1';
        end if;
    end process SUSPENDED;

    STATUS: process
    -- This logic handles several system relevant signals.
    variable CNT_1ms    : integer range 0 to 65535;
    variable LOCK       : boolean;
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;

        if HC_STATE = READ_ITL_PTD or HC_STATE = READ_ATL_PTD then
            ACTUAL_BYTES <= PTD_HEADER(1)(1 downto 0) & PTD_HEADER(0);
        elsif DATA_RDY = '1' then
            ACTUAL_BYTES <= ACTUAL_BYTES + '1';
        end if;
    
        FR_DEC <= '0'; -- Strobe.
        SO <= '0'; -- Strobe.
        
        if HC_STATE /= USB_OPERATIONAL then
            FR_DEC <= Tx_STRB;
        end if;
        
        if HC_STATE /= USB_OPERATIONAL and SOF = '1' then -- Scheduling overrun detected.
            SOC <= SOC + '1';
            SO <= '1';
        end if;
        
        ATL_INT <= '0'; -- ATL interrupt is a strobe;
        
        if HC_STATE = WRITE_ATL_PTD and HC_NSTATE = USB_OPERATIONAL then
            ATL_INT <= '1';
        end if;
        
        UE <= '0'; -- Unrecoverable error, not used currently.
        
        -- Resume is detected when the K state is valid for
        -- 1ms or more.
        
        if DP_IN = '0' or DM_IN = '1' then -- No K state.
            CNT_1ms := 0;
            RD_I <= '0';
            LOCK := false;
        elsif CNT_1ms < 47999 then -- K state detected.
            CNT_1ms := CNT_1ms + 1;
        elsif LOCK = false then -- K state detected.
            RD_I <= '1'; -- Resume detected. RD is a strobe.
            LOCK := true;
        else -- K state detected.
            RD_I <= '0';
        end if;
    end process STATUS;
    
    RD <= RD_I;

    SPEED_Tx <= '1' when HC_STATE <= TOKEN_SOF else
                '1' when HC_STATE <= PREAMBLE else SPEED;

    -- Be aware: the PID is indicated early by SIE_NSTATE for correct shifter loading.
    PID <= "0101" when HC_STATE <= TOKEN_SOF else -- Start of frame.
           "1100" when HC_STATE <= PREAMBLE else -- Low speed preamble.
           "0001" when SIE_NSTATE = TOKEN_PID and DIRECTION_PID = "01" else -- Token out.
           "1001" when SIE_NSTATE = TOKEN_PID and DIRECTION_PID = "10" else -- Token in.
           "1101" when SIE_NSTATE = TOKEN_PID else -- Token setup.
           "0011" when SIE_NSTATE = DATA_PID_Tx and FORMAT = '1' else -- Data0 for ISO data transfers.
           "0011" when SIE_NSTATE = DATA_PID_Tx and TOGGLE = '0' else -- Data0.
           "1011" when SIE_NSTATE = DATA_PID_Tx else -- Data1.
           "0010"; -- Default is the ACK handshake from host to function.

    MAX_PACKET_SIZE <= PTD_HEADER(3)(1 downto 0) & PTD_HEADER(2);
    TOTAL_BYTES <= PTD_HEADER(5)(1 downto 0) & PTD_HEADER(4);

    DATA_TIMING: process
    variable TIMER          : integer range 0 to 31;
    variable REG            : std_logic;
    variable RCV_POSEDGE    : std_logic;
    begin
    wait until CLK_48MHz = '1' and CLK_48MHz' event;
        -- Recever edge detector:
        RCV_POSEDGE := not REG and DP_IN and not DM_IN;
        REG := DP_IN and not DM_IN;

        -- Receive:
        if SIE_STATE /= HANDSHAKE_Rx and SIE_STATE /= DATA_PID_Rx and SIE_STATE /= DATA_Rx then
            Rx_STRB <= '0';
        elsif RCV_POSEDGE = '1' then
            TIMER := 0;
            Rx_STRB <= '0';
        elsif SPEED = '1' then -- Full speed.
            TIMER := TIMER + 1 mod 4;
            case TIMER is -- 1.5Mb/s.
                when 1 => Rx_STRB <= '1';
                when others => Rx_STRB <= '0';
            end case;
        elsif SPEED = '0' then -- Lo speed.
            TIMER := TIMER + 1 mod 32;
            case TIMER is -- 1.5Mb/s.
                when 15 => Rx_STRB <= '1';
                when others => Rx_STRB <= '0';
            end case;
        end if;

        -- Transmit:
        if SPEED = '1' then -- Full speed.
            TIMER := TIMER + 1 mod 4;
            case TIMER is -- 12Mb/s.
                when 1 => Tx_STRB <= '1';
                when others => Tx_STRB <= '0';
            end case;
        else
            TIMER := TIMER + 1 mod 32;
            case TIMER is -- 1.5Mb/s.
                when 15 => Tx_STRB <= '1';
                when others => Tx_STRB <= '0';
            end case;
        end if;
    end process DATA_TIMING;

    P_HUBSETUP: process
    -- This counter controls the setup time of 4 high
    -- speed clock cycles after a low speed preamble 
    -- has been sent. So we have to count 16 48MHz
    -- clock cycles.
    variable TIMER      : integer range 0 to 15;
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if SIE_STATE /= WAIT_HUBSETUP then
            SETUP_RDY <= '0';
            TIMER := 0;
        elsif TIMER < 15 then
            TIMER := TIMER + 1;
        else
            SETUP_RDY <= '1';
        end if;
    end process P_HUBSETUP;

    SIE_STATEREG: process
    -- This is the serial interface engine (SIE) 
    -- state machines register.
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if HC_STATE = USB_RESET or HC_STATE = USB_SUSPEND then
            SIE_STATE <= IDLE;
        elsif HCR = '1' then
            SIE_STATE <= IDLE;
        else
            SIE_STATE <= SIE_NSTATE;
        end if;
    end process SIE_STATEREG;

    SIE_STATEDEC: process(BUBBLE_LOA, CRC_ERROR, HC_STATE, MAX_PACKET_SIZE, PID, Rx_RDY, Rx_HOLD_REG, SETUP_RDY, SIE_STATE, TIMEOUT, Tx_RDY, Tx_STRB, WRONG_EOP)
    -- This is the next state decoder of the serial interface engine (SIE) 
    -- for the flow control of the four USB transactions (bulk, control, 
    -- interrupt and isochronous).
    begin
        case SIE_STATE is
            when IDLE =>
                if Tx_STRB = '1' and (HC_STATE = TOKEN_SOF or HC_STATE = PREAMBLE or HC_STATE = TRANSMIT_ITL_DATA or 
                                      HC_STATE = RECEIVE_ITL_DATA or HC_STATE = TRANSMIT_ATL_DATA or HC_STATE = RECEIVE_ATL_DATA) then
                    SIE_NSTATE <= SYNC;
                else
                    SIE_NSTATE <= IDLE;
                end if;
            when SYNC => -- Send Sync pattern.
                if Tx_RDY = '1' then
                    SIE_NSTATE <= TOKEN_PID;
                else
                    SIE_NSTATE <= SYNC;
                end if;
            when TOKEN_PID =>
                if Tx_RDY = '1' and PID = "1100" then -- Preamble.
                    SIE_NSTATE <= WAIT_HUBSETUP;
                elsif Tx_RDY = '1' then
                    SIE_NSTATE <= TOKEN_Tx;
                else
                    SIE_NSTATE <= TOKEN_PID;
                end if;
            when WAIT_HUBSETUP =>
                if SETUP_RDY = '1' then
                    SIE_NSTATE <= IDLE;
                else
                    SIE_NSTATE <= WAIT_HUBSETUP;
                end if;
            when TOKEN_Tx => 
                if Tx_RDY = '1' then
                    SIE_NSTATE <= TOKEN_EOP;
                else
                    SIE_NSTATE <= TOKEN_Tx;
                end if;
            when TOKEN_EOP =>
                if Tx_RDY = '1' and (HC_STATE = RECEIVE_ITL_DATA or HC_STATE = RECEIVE_ATL_DATA) then
                    SIE_NSTATE <= DATA_PID_Rx;
                elsif Tx_RDY = '1' and HC_STATE = TRANSMIT_ITL_DATA and MAX_PACKET_SIZE = "0000000000" then
                    SIE_NSTATE <= IDLE; -- No ISO data: transfer finished.
                elsif Tx_RDY = '1' and MAX_PACKET_SIZE = "0000000000" then
                    SIE_NSTATE <= HANDSHAKE_Rx; -- No control or bulk data.
                elsif Tx_RDY = '1' then
                    SIE_NSTATE <= SYNC_DATA;
                else
                    SIE_NSTATE <= TOKEN_EOP;
                end if;
            when SYNC_DATA =>
                if Tx_RDY = '1' then
                    SIE_NSTATE <= DATA_PID_Tx;
                else
                    SIE_NSTATE <= SYNC_DATA;
                end if;
            when DATA_PID_Tx =>
                if Tx_RDY = '1' then
                    SIE_NSTATE <= DATA_Tx;
                else
                    SIE_NSTATE <= DATA_PID_Tx;
                end if;
            when DATA_Tx =>
                if Tx_RDY = '1' then
                    SIE_NSTATE <= DATA_EOP;
                else
                    SIE_NSTATE <= DATA_Tx;
                end if;
            when DATA_EOP =>
                if Tx_RDY = '1' and HC_STATE = TRANSMIT_ITL_DATA then
                    SIE_NSTATE <= IDLE; -- ISO has no handshake.
                elsif Tx_RDY = '1' then
                    SIE_NSTATE <= HANDSHAKE_Rx;
                else
                    SIE_NSTATE <= DATA_EOP;
                end if;
            when HANDSHAKE_Rx =>
                if Rx_RDY = '1' or TIMEOUT = '1' or WRONG_EOP = '1' or BUBBLE_LOA = '1' then
                    SIE_NSTATE <= IDLE;
                else
                    SIE_NSTATE <= HANDSHAKE_Rx;
                end if;
            when DATA_PID_Rx =>
                if TIMEOUT = '1' or BUBBLE_LOA = '1' then
                    SIE_NSTATE <= IDLE; -- Finish transaction.
                elsif Rx_RDY = '1' and Rx_HOLD_REG = x"C3" and FORMAT = '0' and TOGGLE = '1' then -- DATA0 PID.
                    SIE_NSTATE <= IDLE; -- Wrong PID. Expected 1, received 0.
                elsif Rx_RDY = '1' and Rx_HOLD_REG = x"4B" and FORMAT = '0' and TOGGLE = '0' then -- DATA1 PID.
                    SIE_NSTATE <= IDLE; -- Wrong PID. Expected 0 received 1.
                elsif Rx_RDY = '1' and Rx_HOLD_REG = x"C3" then -- DATA0 PID.
                    SIE_NSTATE <= DATA_Rx;
                elsif Rx_RDY = '1' and Rx_HOLD_REG = x"4B" then -- DATA1 PID.
                    SIE_NSTATE <= DATA_Rx;
                elsif Rx_RDY = '1' then -- NAK, STALL.
                    SIE_NSTATE <= IDLE;
                else
                    SIE_NSTATE <= DATA_PID_Rx;
                end if;
            when DATA_Rx =>
                if CRC_ERROR = '1' or WRONG_EOP = '1' or BUBBLE_LOA = '1' then
                    SIE_NSTATE <= IDLE; -- Finish transaction.
                elsif Rx_RDY = '1' and HC_STATE = RECEIVE_ITL_DATA then
                    SIE_NSTATE <= IDLE; -- ISO has no handshake.
                elsif Rx_RDY = '1' then
                    SIE_NSTATE <= HANDSHAKE_Tx;
                else
                    SIE_NSTATE <= DATA_Rx;
                end if;
            when HANDSHAKE_Tx =>
                if Tx_RDY = '1' then
                    SIE_NSTATE <= IDLE;
                else
                    SIE_NSTATE <= HANDSHAKE_Tx;
                end if;
        end case;
    end process SIE_STATEDEC;

    with SIE_STATE select
        DPM_EN <= '1' when SYNC | TOKEN_PID | TOKEN_Tx | TOKEN_EOP | SYNC_DATA | DATA_PID_Tx | DATA_Tx | DATA_EOP | HANDSHAKE_Tx,
                  '0' when others;

    NRZI_TRANSCEIVER: process
    -- This is the USB transceiver. The data is NRZI (Non Return to Zero Inverted)
    -- encoded and decoded. This transceiver features bit stuffing. The USB data is
    -- sent LSB first to the USB bus. The shift register is arranged accordingly.
    -- The transmission is controlled by the Tx_STRB strobe After each byte sent,
    -- the Rx_RDY strobe indicates the requirement for the next byte to be sent.
    -- After each byte received, the Rx_RDY strobe indicates valid received data.
    variable BITCNT         : integer range 0 to 15;
    variable SHIFTER        : std_logic_vector(7 downto 0);
    variable STUFF_CNT      : std_logic_vector(2 downto 0); -- Bit stuffing: counts '1's.
    variable NRZI_OUT       : std_logic;
    variable NRZI_IN        : std_logic;
    variable USB_D_IN       : std_logic;
    variable USB_D_OUT      : std_logic;
    variable CRC5           : std_logic_vector(5 downto 1);
    variable CRC16          : std_logic_vector(16 downto 1);
    variable CRC16_HI       : boolean;
    variable CRC16_LO       : boolean;
    variable Rx_ACTIVE      : std_logic;
    variable EOP_BIT        : integer range 0 to 3;
    variable PACKET_SIZE    : std_logic_vector(9 downto 0);
    variable TIMEOUT_CNT    : std_logic_vector(4 downto 0);
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;

        DATA_RDY <= '0'; -- Strobe.
        Rx_RDY <= '0';  -- Strobe.
        Tx_RDY <= '0';  -- Strobe.
        CRC_ERROR <= '0'; -- Strobe;
        STUFF_ERR <= '0'; -- Strobe.
        WRONG_EOP <= '0'; -- Strobe.
        BUBBLE_LOA <= '0'; -- Strobe.
        OPBERR <= '0'; -- Strobe.
        

        if SIE_STATE = IDLE then
            BITCNT := 0;
            STUFF_CNT := "000";
            NRZI_OUT := '0';
            NRZI_IN := '0';
            USB_D_IN := '1'; -- Idle.
            USB_D_OUT := '1'; -- Idle.
        else
            -- Receiver data:
            USB_D_IN := USB_D_IN xnor NRZI_IN; -- NRZI decoding.
            NRZI_IN := DP_IN and not DM_IN;
        end if;
        
        -- Transmitter initialization:
        if SIE_STATE /= SYNC and SIE_NSTATE = SYNC then
            SHIFTER := x"80"; -- This is the sync pattern.
        elsif SIE_STATE /= TOKEN_PID and SIE_NSTATE = TOKEN_PID then
            SHIFTER := not PID & PID;
        elsif SIE_STATE /= TOKEN_Tx and SIE_NSTATE = TOKEN_Tx and PID = "0101" then -- Token start of frame.
            SHIFTER := FRAME_NUMBER(7 downto 0);
        elsif SIE_STATE /= TOKEN_Tx and SIE_NSTATE = TOKEN_Tx and PID = "0001" then -- Token: in out or setup.
            SHIFTER := ENDPOINT_NUMBER(0) & FUNCTION_ADDRESS;
        elsif SIE_STATE /= SYNC_DATA and SIE_NSTATE = SYNC_DATA then
            SHIFTER := x"80"; -- This is the sync pattern.
        elsif SIE_STATE /= DATA_PID_Tx and SIE_NSTATE = DATA_PID_Tx then
            SHIFTER := not PID & PID;
        elsif SIE_STATE /= DATA_Tx and SIE_NSTATE = DATA_Tx then
            SHIFTER := BUFFER_IN(7 downto 0); -- This is the first byte to be sent.
            PACKET_SIZE := "0000000000";
            CRC16_HI := false;
            CRC16_LO := false;
        elsif SIE_STATE /= HANDSHAKE_Tx and SIE_NSTATE = HANDSHAKE_Tx then
            SHIFTER := x"D2"; -- This is ACK.
        end if;
        
        -- Transmit:
        if Tx_STRB = '1' and (SIE_STATE = SYNC or SIE_STATE = TOKEN_PID or SIE_STATE = TOKEN_Tx or 
                              SIE_STATE = SYNC_DATA or SIE_STATE = DATA_PID_Tx or SIE_STATE = DATA_Tx or SIE_STATE = HANDSHAKE_Tx)then
            if STUFF_CNT = "110" then -- Stuff a zero after six detected data "ones".
                STUFF_CNT := "000";
                USB_D_OUT := '0';
            elsif (SIE_STATE = SYNC or SIE_STATE = TOKEN_PID or SIE_STATE = SYNC_DATA or SIE_STATE = DATA_PID_Tx or SIE_STATE = HANDSHAKE_Tx) and BITCNT = 7 then
                Tx_RDY <= '1';
                BITCNT := 0;
                USB_D_OUT := SHIFTER(0); -- LSB first.
                STUFF_CNT := STUFF_CNT + SHIFTER(0); -- Count '1's.
            elsif SIE_STATE = TOKEN_Tx and BITCNT = 7 then
                case PID is
                    when "0101" => SHIFTER := "00000" & FRAME_NUMBER(10 downto 8); -- TOKEN_SOF.
                    when others => SHIFTER := "00000" & ENDPOINT_NUMBER(3 downto 1);
                end case;            
                USB_D_OUT := SHIFTER(0); -- LSB first.
                STUFF_CNT := STUFF_CNT + SHIFTER(0); -- Count '1's.
            elsif SIE_STATE = TOKEN_Tx and BITCNT = 10 then
                SHIFTER := "000" & CRC5;
                USB_D_OUT := SHIFTER(0); -- LSB first.
                STUFF_CNT := STUFF_CNT + SHIFTER(0); -- Count '1's.
            elsif SIE_STATE = TOKEN_Tx and BITCNT = 15 then
                Tx_RDY <= '1';
                BITCNT := 0;
                USB_D_OUT := SHIFTER(0); -- LSB first.
                STUFF_CNT := STUFF_CNT + SHIFTER(0); -- Count '1's.
            elsif SIE_STATE = DATA_Tx and BITCNT = 7 then
                if CRC16_HI = true then
                    Tx_RDY <= '1';
                elsif CRC16_LO = true then
                    SHIFTER := CRC16(16 downto 9);
                    CRC16_HI := true;
                elsif PACKET_SIZE = MAX_PACKET_SIZE or ACTUAL_BYTES = TOTAL_BYTES then
                    SHIFTER := CRC16(8 downto 1);
                    CRC16_LO := true;
                else
                    SHIFTER := BUFFER_IN(15 downto 8);
                    PACKET_SIZE := PACKET_SIZE + '1';
                    DATA_RDY <= '1';
                end if;
                USB_D_OUT := SHIFTER(0); -- LSB first.
                STUFF_CNT := STUFF_CNT + SHIFTER(0); -- Count '1's.
            elsif SIE_STATE = DATA_Tx and BITCNT = 15 then
                if CRC16_HI = true then
                    Tx_RDY <= '1';
                elsif CRC16_LO = true then
                    SHIFTER := CRC16(16 downto 9);
                    CRC16_HI := true;
                elsif PACKET_SIZE = MAX_PACKET_SIZE or ACTUAL_BYTES = TOTAL_BYTES then
                    SHIFTER := CRC16(8 downto 1);
                    CRC16_LO := true;
                else
                    SHIFTER := BUFFER_IN(7 downto 0);
                    PACKET_SIZE := PACKET_SIZE + '1';
                    DATA_RDY <= '1';
                end if;

                BITCNT := 0;
                USB_D_OUT := SHIFTER(0); -- LSB first.
                STUFF_CNT := STUFF_CNT + SHIFTER(0); -- Count '1's.
            else
                BITCNT := BITCNT + 1;
                SHIFTER := '0' & SHIFTER(7 downto 1);
                USB_D_OUT := SHIFTER(0); -- LSB first.
                STUFF_CNT := STUFF_CNT + SHIFTER(0); -- Count '1's.
            end if;

            NRZI_OUT := NRZI_OUT xor USB_D_OUT; -- NRZI encoding.
            DP_OUT <= NRZI_OUT;
            DM_OUT <= not NRZI_OUT;
        elsif SIE_STATE = TOKEN_EOP or SIE_STATE = DATA_EOP then
            if Tx_STRB = '1' then
                BITCNT := BITCNT + 1;
            end if;
            
            case BITCNT is
                when 0 => -- Send SE0.
                    DP_OUT <= '0';
                    DM_OUT <= '0';
                when 1 => -- Send J in full speed.
                    DP_OUT <= '1';
                    DM_OUT <= '0';
                when 2 => -- Send J in lo speed.
                    DP_OUT <= '0';
                    DM_OUT <= '1';
                when others =>
                    Tx_RDY <= '1';
            end case;
        end if;

        -- Receiver initialization:
        if SIE_STATE /= DATA_PID_Rx and SIE_NSTATE = DATA_PID_Rx then
            SHIFTER := x"00"; -- Clear.
            Rx_ACTIVE := '0';
            EOP_BIT := 0;
        elsif SIE_STATE /= DATA_Rx and SIE_NSTATE = DATA_Rx then
            SHIFTER := x"00"; -- Clear.
        end if;

        -- Receieve:
        -- This logic controls the incoming data stream. The data is valid when
        -- the sync pattern has been detected and is valid until the EOP pattern
        -- is detected by this logic.
        if STUFF_CNT = "111" and Rx_ACTIVE = '1' then
            if USB_D_IN = '1' then
                STUFF_ERR <= '1'; -- Bit stuffing error.
            end if;
            STUFF_CNT := "000";
            -- Abandon the stuffed zero.
        elsif (SIE_STATE = DATA_PID_Rx or SIE_STATE = HANDSHAKE_Rx) and Rx_STRB = '1' and BITCNT = 7 then
            BITCNT := 0;
            SHIFTER := USB_D_IN & SHIFTER(7 downto 1); -- LSB comes first.
            STUFF_CNT := STUFF_CNT + USB_D_IN; -- Count '1's.
            --
            if Rx_ACTIVE = '1' then
                Rx_HOLD_REG(7 downto 0) <= SHIFTER;
                Rx_RDY <= '1';
            end if;
            --
            if SHIFTER = x"80" then
                Rx_ACTIVE := '1'; -- Sync pattern detected.
            end if;
        elsif SIE_STATE = DATA_Rx and Rx_STRB = '1' and BITCNT = 7 then
            BITCNT := 0;
            SHIFTER := USB_D_IN & SHIFTER(7 downto 1); -- LSB comes first.
            STUFF_CNT := STUFF_CNT + USB_D_IN; -- Count '1's.
            --
            if Rx_ACTIVE = '0' and EOP_BIT = 3 then
                Rx_RDY <= '1';
                if CRC16 /= x"0000" then
                    CRC_ERROR <= '1';
                end if;    
            elsif Rx_ACTIVE = '1' then
                Rx_HOLD_REG(7 downto 0) <= SHIFTER;
                DATA_RDY <= '1';
            end if;
        elsif SIE_STATE = DATA_Rx and Rx_STRB = '1' and BITCNT = 15 then
            BITCNT := 0;
            SHIFTER := USB_D_IN & SHIFTER(7 downto 1); -- LSB comes first.
            STUFF_CNT := STUFF_CNT + USB_D_IN; -- Count '1's.
            --
            if Rx_ACTIVE = '0' and EOP_BIT = 3 then
                Rx_RDY <= '1';
                if CRC16 /= x"0000" then
                    CRC_ERROR <= '1';
                end if;    
            elsif Rx_ACTIVE = '1' then
                Rx_HOLD_REG(7 downto 0) <= SHIFTER;
                DATA_RDY <= '1';
            end if;
        else
            if Rx_ACTIVE = '0' and EOP_BIT = 3 then
                Rx_RDY <= '1';
            end if;
            --
            BITCNT := BITCNT + 1;
            SHIFTER := USB_D_IN & SHIFTER(7 downto 1); -- LSB comes first.
            STUFF_CNT := STUFF_CNT + USB_D_IN; -- Count '1's.
        end if;

        -- This logic controls the incoming data stream. The data is valid when
        -- the sync pattern has been detected and is valid until the EOP pattern
        -- is detected by this logic.
        if SIE_STATE = DATA_PID_Rx or SIE_STATE = DATA_Rx or SIE_STATE = HANDSHAKE_Rx then
            if DP_IN = '0' and DM_IN = '0' and EOP_BIT = 0 and Rx_STRB = '1' then -- Fisrt SE0.
                EOP_BIT := 1;
                Rx_ACTIVE := '0';
            elsif (DP_IN xor DM_IN) = '1' and EOP_BIT = 1 and Rx_STRB = '1' then
                WRONG_EOP <= '1';
                OPBERR <= '1';
            elsif DP_IN = '0' and DM_IN = '0' and EOP_BIT = 1 and Rx_STRB = '1' then -- Second SE0.
                EOP_BIT := 2;
            elsif (DP_IN /= '1' or DM_IN /= '0') and EOP_BIT = 2 and Rx_STRB = '1' then
                WRONG_EOP <= '1';
                OPBERR <= '1';
            elsif DP_IN = '1' and DM_IN = '0' and EOP_BIT = 2 and Rx_STRB = '1' then -- Full speed J.
                EOP_BIT := 3;
            end if;
        end if;

        if SOF = '1' and Rx_ACTIVE = '1' then
            BUBBLE_LOA <= '1';
            OPBERR <= '1'; -- Bubbling device or loss of activity.
        end if;

        -- This is the cyclic redundancy logic for CRC5 and CRC16
        -- generation and validation.
        if SIE_STATE = IDLE then
            CRC5 := "00000";
            CRC16 := x"0000";
        elsif SIE_STATE = TOKEN_Tx and Tx_STRB = '1' and BITCNT < 11 then
            -- The CRC5 polynomial is G(x) = x^5 + x^2 + 1.
            CRC5 := CRC5(4 downto 3) & (CRC5(2) xor CRC5(5) xor SHIFTER(0)) & CRC5(1) & (CRC5(5) xor SHIFTER(0));
        elsif SIE_STATE = DATA_Tx and (PACKET_SIZE = MAX_PACKET_SIZE or ACTUAL_BYTES = TOTAL_BYTES) then
            null; -- Do not shift further.
        elsif SIE_STATE = DATA_Tx and Tx_STRB = '1' then
            -- The CRC16 polynomial is G(x) = x^16 + x^15 + x^2 + 1.
            CRC16 := (CRC16(15) xor CRC16(16) xor SHIFTER(0)) & CRC16(14 downto 3) & (CRC16(2) xor CRC16(16) xor SHIFTER(0)) & CRC16(1) & (CRC16(16) xor SHIFTER(0));
        elsif SIE_STATE = DATA_Rx and Rx_STRB = '1' and Rx_ACTIVE = '1' then
            -- The CRC16 polynomial is G(x) = x^16 + x^15 + x^2 + 1.
            CRC16 := (CRC16(15) xor CRC16(16) xor SHIFTER(0)) & CRC16(14 downto 3) & (CRC16(2) xor CRC16(16) xor SHIFTER(0)) & CRC16(1) & (CRC16(16) xor SHIFTER(0));
        end if;

        -- Receiver timeout:
        if SIE_STATE /= DATA_PID_Rx and SIE_STATE /= DATA_Rx and SIE_STATE /= HANDSHAKE_Rx then
            TIMEOUT <= '0';
            TIMEOUT_CNT := "00000";
        elsif Rx_ACTIVE = '1' then
            TIMEOUT_CNT := "00000";
        elsif Rx_STRB = '1' and TIMEOUT_CNT < "10001" then
            TIMEOUT_CNT := TIMEOUT_CNT + '1';
        elsif Rx_STRB = '1' then -- 18 bit times.
            TIMEOUT <= '1';
            Rx_HOLD_REG <= x"0000";
        end if;
    end process NRZI_TRANSCEIVER;
end architecture BEHAVIOUR;
