------------------------------------------------------------------------
----                                                                ----
---- WF1160 IP Core                                                 ----
----                                                                ----
---- Description:                                                   ----
---- This model provides an embedded Universal Serial Bus host      ----
---- controller compatible to the Philips ISP1160.                  ----
----                                                                ----
---- This  is the top level file with tree state buses.             ----
----                                                                ----
----                                                                ----
----                                                                ----
----                                                                ----
---- Author(s):                                                     ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de     ----
----                                                                ----
------------------------------------------------------------------------
----                                                                ----
---- Copyright  2020... Wolfgang Foerster - Inventronik GmbH.      ----
----                                                                ----
---- This source file may be used and distributed without           ----
---- restriction provided that this copyright statement is not      ----
---- removed from the file and that any derivative work contains    ----
---- the original copyright notice and the associated disclaimer.   ----
----                                                                ----
---- This source file is free software; you can redistribute it     ----
---- and/or modify it under the terms of the GNU Lesser General     ----
---- Public License as published by the Free Software Foundation;   ----
---- either version 2.1 of the License, or (at your option) any     ----
---- later version.                                                 ----
----                                                                ----
---- This source is distributed in the hope that it will be         ----
---- useful, but WITHOUT ANY WARRANTY; without even the implied     ----
---- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR        ----
---- PURPOSE. See the GNU Lesser General Public License for more    ----
---- details.                                                       ----
----                                                                ----
---- You should have received a copy of the GNU Lesser General      ----
---- Public License along with this source; if not, download it     ----
---- from http://www.gnu.org/licenses/lgpl.html                     ----
----                                                                ----
------------------------------------------------------------------------
--
-- Revision History
--
-- Revision 2K20B  20201224 WF
--   Initial Release.
--

library work;
use work.wf1160_pkg.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity USB1160_TOP is
    port (
        -- System controls:
        CLK_48MHz   : in std_logic;
        RESETn      : in std_logic;

        -- Address and data:
        A0          : in std_logic;
        DATA_IN     : in  std_logic_vector(15 downto 0);
        DATA_OUT    : out std_logic_vector(15 downto 0);
        DATA_EN     : out std_logic;

        -- Bus controls:
        CSn         : in std_logic; -- Chip select.
        RDn         : in std_logic; -- Read data.
        WRn         : in std_logic; -- Write data.
        EOT         : in std_logic; -- End of DMA Transfer.
        DACKn       : in std_logic; -- DMA data acknowledge.
        DREQ        : out std_logic; -- DMA data request.
        INT         : out std_logic; -- Interrupt.

        -- USB host:
        WAKEUP      : in std_logic; -- Wakeup from suspend.
        SUSPEND     : out std_logic; -- Suspend status.
        NDP_SEL     : in std_logic; -- Number of data ports.
        PSW1n       : out std_logic; -- Power switch.
        PSW2n       : out std_logic; -- Power switch.
        OC1n        : in std_logic; -- Overcurrent detection.
        OC2n        : in std_logic; -- Overcurrent detection.
        DM1         : inout std_logic; -- Data line.
        DP1         : inout std_logic; -- Data line.
        DM2         : inout std_logic; -- Data line.
        DP2         : inout std_logic; -- Data line.
        DP15K       : out std_logic -- Switch for four 15K pull down resistors
    );
end entity USB1160_TOP;

architecture STRUCTURE of USB1160_TOP is
signal ATL_BUFF_DONE        : std_logic;
signal ATL_BUFF_EMPTY       : std_logic;
signal ATL_BUFF_FULL        : std_logic;
signal ATL_BUFF_LEN         : std_logic_vector(11 downto 0);
signal ATL_INT              : std_logic;
signal ATL_RD_HC            : std_logic;
signal ATL_RD_MC            : std_logic;
signal ATL_WR_HC            : std_logic;
signal ATL_WR_MC            : std_logic;
signal AOCEN                : std_logic;
signal DATA_IN_MC           : std_logic_vector(15 downto 0);
signal DATA_OUT_MC          : std_logic_vector(15 downto 0);
signal BUFFER_IN_HC         : std_logic_vector(15 downto 0);
signal BUFFER_IN_MC         : std_logic_vector(15 downto 0);
signal BUFFER_OUT_HC        : std_logic_vector(15 downto 0);
signal BUFFER_OUT_MC        : std_logic_vector(15 downto 0);
signal CCS_1                : std_logic;
signal CCS_2                : std_logic;
signal CLKNS                : std_logic;
signal DP1_OUT              : std_logic;
signal DM1_OUT              : std_logic;
signal DPM1_EN              : std_logic;
signal DP2_OUT              : std_logic;
signal DM2_OUT              : std_logic;
signal DPM2_EN              : std_logic;
signal DM_IN                : std_logic;
signal DP_IN                : std_logic;
signal DM_OUT               : std_logic;
signal DP_OUT               : std_logic;
signal DPM_EN               : std_logic;
signal HC_RESUME            : std_logic;
signal FR                   : std_logic_vector(13 downto 0);
signal ATL_FIFO_FULL        : std_logic;
signal ITL1_FIFO_FULL       : std_logic;
signal ITL0_FIFO_FULL       : std_logic;
signal ATL_FIFO_EMPTY       : std_logic;
signal ITL1_FIFO_EMPTY      : std_logic;
signal ITL0_FIFO_EMPTY      : std_logic;
signal FR_DEC               : std_logic;
signal FSMPS                : std_logic_vector(14 downto 0);
signal HC_ADR               : std_logic_vector(11 downto 0);
signal HC_ITL1              : std_logic;
signal HC_SUSPND            : std_logic;
signal HCFS_HC              : std_logic_vector(1 downto 0);
signal HCFS_MC              : std_logic_vector(1 downto 0);
signal HCR                  : std_logic;
signal ITL0_BUFF_DONE       : std_logic;
signal ITL1_BUFF_DONE       : std_logic;
signal ITL0_BUFF_EMPTY      : std_logic;
signal ITL1_BUFF_EMPTY      : std_logic;
signal ITL0_BUFF_FULL       : std_logic;
signal ITL1_BUFF_FULL       : std_logic;
signal RD_ITL0_BUFF_LENGTH  : std_logic_vector(15 downto 0);
signal RD_ITL1_BUFF_LENGTH  : std_logic_vector(15 downto 0);
signal ITL_BUFF_LEN         : std_logic_vector(11 downto 0);
signal ITL_RD_HC            : std_logic;
signal ITL_RD_MC            : std_logic;
signal ITL_WR_HC            : std_logic;
signal ITL_WR_MC            : std_logic;
signal LSDA_1               : std_logic;
signal LSDA_2               : std_logic;
signal LST                  : std_logic_vector(10 downto 0);
signal OPBERR               : std_logic;
signal OPBERR_1             : std_logic;
signal OPBERR_2             : std_logic;
signal PESC_1               : std_logic;
signal PESC_2               : std_logic;
signal POCI_1               : std_logic;
signal POCI_2               : std_logic;
signal PPS_1                : std_logic;
signal PPS_2                : std_logic;
signal PRS_1                : std_logic;
signal PRS_2                : std_logic;
signal PRSC_1               : std_logic;
signal PRSC_2               : std_logic;
signal PSSC_1               : std_logic;
signal PSSC_2               : std_logic;
signal RD                   : std_logic;
signal RESET                : std_logic;
signal RHSC                 : std_logic;
signal RWE                  : std_logic;
signal SF                   : std_logic;
signal SO                   : std_logic;
signal SOC                  : std_logic_vector(1 downto 0);
signal SOFTWARE_RESET       : std_logic;
signal SPEED_Tx             : std_logic;
signal UPSTREAM_P1_P2n      : std_logic;
signal UE                   : std_logic;
begin
    SYNC: process
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        RESET <= not RESETn;
    end process SYNC;

    -- At this point we switch the little endian format to the big endian.
    -- The USB1160 is little endian and is internally big endian.
    DATA_IN_MC <= DATA_IN(7 downto 0) & DATA_IN(15 downto 8);
    DATA_OUT <= DATA_OUT_MC(7 downto 0) & DATA_OUT_MC(15 downto 8);

    DM1 <= DM1_OUT when DPM1_EN = '1' else 'Z';
    DP1 <= DP1_OUT when DPM1_EN = '1' else 'Z';

    DM2 <= DM2_OUT when DPM2_EN = '1' else 'Z';
    DP2 <= DP2_OUT when DPM2_EN = '1' else 'Z';

    OPBERR_1 <= OPBERR when UPSTREAM_P1_P2n = '1' else '0';
    OPBERR_2 <= OPBERR when UPSTREAM_P1_P2n = '0' else '0';

    I_MC_IF: MC_INTERFACE
        port map(
            CLK_48MHz               => CLK_48MHz,
            RESET                   => RESET,
            A0                      => A0,
            CSn                     => CSn,
            RDn                     => RDn,
            WRn                     => WRn,
            DATA_IN                 => DATA_IN_MC,
            DATA_OUT                => DATA_OUT_MC,
            DATA_EN                 => DATA_EN,
            NDP_SEL                 => NDP_SEL,
            RWE                     => RWE,
            HC_RESUME               => HC_RESUME,
            HCFS_IN                 => HCFS_HC,
            HCFS_OUT                => HCFS_MC,
            SOC                     => SOC,
            HCR                     => HCR,
            SOFTWARE_RESET          => SOFTWARE_RESET,
            RHSC                    => RHSC,
            UE                      => UE,
            RD                      => RD,
            SF                      => SF,
            SO                      => SO,
            FSMPS                   => FSMPS,
            FR                      => FR,
            FR_DEC                  => FR_DEC,
            LST                     => LST,
            CCS_1                   => CCS_1,
            LSDA_1                  => LSDA_1,
            OPBERR_1                => OPBERR_1,
            PESC_1                  => PESC_1,
            POCI_1                  => POCI_1,
            PRSC_1                  => PRSC_1,
            PSSC_1                  => PSSC_1,
            PPS_1                   => PPS_1,
            PRS_1                   => PRS_1,
            CCS_2                   => CCS_2,
            LSDA_2                  => LSDA_2,
            OPBERR_2                => OPBERR_2,
            PESC_2                  => PESC_2,
            POCI_2                  => POCI_2,
            PRSC_2                  => PRSC_2,
            PSSC_2                  => PSSC_2,
            PPS_2                   => PPS_2,
            PRS_2                   => PRS_2,
            DP15K                   => DP15K,
            CLKNS                   => CLKNS,
            AOCEN                   => AOCEN,
            INT                     => INT,
            DACKn                   => DACKn,
            DREQ                    => DREQ,
            EOT                     => EOT,
            CLK_RDY                 => '1', -- Not used.
            HC_SUSPND               => HC_SUSPND,
            ATL_INT                 => ATL_INT,
            ITL_BUFF_LEN            => ITL_BUFF_LEN,
            ATL_BUFF_LEN            => ATL_BUFF_LEN,
            ATL_BUFF_DONE           => ATL_BUFF_DONE,
            ITL1_BUFF_DONE          => ITL1_BUFF_DONE,
            ITL0_BUFF_DONE          => ITL0_BUFF_DONE,
            ATL_BUFF_FULL           => ATL_BUFF_FULL,
            ITL1_BUFF_FULL          => ITL1_BUFF_FULL,
            ITL0_BUFF_FULL          => ITL0_BUFF_FULL,
            RD_ITL0_BUFF_LENGTH     => RD_ITL0_BUFF_LENGTH,
            RD_ITL1_BUFF_LENGTH     => RD_ITL1_BUFF_LENGTH,
            BUFFER_IN               => BUFFER_IN_MC,
            BUFFER_OUT              => BUFFER_OUT_MC,
            ITL_RD                  => ITL_RD_MC,
            ITL_WR                  => ITL_WR_MC,
            ATL_RD                  => ATL_RD_MC,
            ATL_WR                  => ATL_WR_MC,
            HC_ITL1                 => HC_ITL1
        );
        
    I_FIFO: FIFO_BUFFER
        port map(
            CLK_48MHz               => CLK_48MHz,
            RESET                   => RESET,

            ITL_BUFF_LEN            => ITL_BUFF_LEN,
            ATL_BUFF_LEN            => ATL_BUFF_LEN,
            RD_ITL0_BUFF_LENGTH     => RD_ITL0_BUFF_LENGTH,
            RD_ITL1_BUFF_LENGTH     => RD_ITL1_BUFF_LENGTH,

            ATL_FIFO_FULL           => ATL_FIFO_FULL,
            ITL1_FIFO_FULL          => ITL1_FIFO_FULL,
            ITL0_FIFO_FULL          => ITL0_FIFO_FULL,
            ATL_FIFO_EMPTY          => ATL_FIFO_EMPTY,
            ITL1_FIFO_EMPTY         => ITL1_FIFO_EMPTY,
            ITL0_FIFO_EMPTY         => ITL0_FIFO_EMPTY,

            BUFFER_IN_MC            => BUFFER_OUT_MC,
            BUFFER_IN_HC            => BUFFER_OUT_HC,
            BUFFER_OUT_MC           => BUFFER_IN_MC,
            BUFFER_OUT_HC           => BUFFER_IN_HC,

            ATL_RD_MC               => ATL_RD_MC,
            ATL_WR_MC               => ATL_WR_MC,
            ITL_RD_MC               => ITL_RD_MC,
            ITL_WR_MC               => ITL_WR_MC,

            ATL_RD_HC               => ATL_RD_HC,
            ATL_WR_HC               => ATL_WR_HC,
            ITL_RD_HC               => ITL_RD_HC,
            ITL_WR_HC               => ITL_WR_HC,
            HC_ITL1                 => HC_ITL1,
            HC_ADR                  => HC_ADR
        );

    I_HC: HOST_CONTROLLER
        port map(
            CLK_48MHz               => CLK_48MHz,
            RESET                   => RESET,

            CSn                     => CSn,

            DP_IN                   => DP_OUT,
            DM_IN                   => DM_OUT,
            DP_OUT                  => DP_IN, 
            DM_OUT                  => DM_IN, 
            DPM_EN                  => DPM_EN,
            SPEED_Tx                => SPEED_Tx,

            RWE                     => RWE,
            HC_RESUME               => HC_RESUME,
            WAKEUP                  => WAKEUP,
            SUSPEND                 => SUSPEND,

            ITL_BUFF_LEN            => ITL_BUFF_LEN,
            ATL_BUFF_LEN            => ATL_BUFF_LEN,
            BUFFER_IN               => BUFFER_IN_HC,
            BUFFER_OUT              => BUFFER_OUT_HC,

            ATL_BUFF_FULL           => ATL_BUFF_FULL,
            ITL0_BUFF_FULL          => ITL0_BUFF_FULL,
            ITL1_BUFF_FULL          => ITL1_BUFF_FULL,
            ATL_BUFF_DONE           => ATL_BUFF_DONE,
            ITL1_BUFF_DONE          => ITL1_BUFF_DONE,
            ITL0_BUFF_DONE          => ITL0_BUFF_DONE,

            ATL_FIFO_FULL           => ATL_FIFO_FULL,
            ITL1_FIFO_FULL          => ITL1_FIFO_FULL,
            ITL0_FIFO_FULL          => ITL0_FIFO_FULL,
            ATL_FIFO_EMPTY          => ATL_FIFO_EMPTY,
            ITL1_FIFO_EMPTY         => ITL1_FIFO_EMPTY,
            ITL0_FIFO_EMPTY         => ITL0_FIFO_EMPTY,

            ATL_RD                  => ATL_RD_HC,
            ATL_WR                  => ATL_WR_HC,
            ITL_RD                  => ITL_RD_HC,
            ITL_WR                  => ITL_WR_HC,
            HC_ITL1                 => HC_ITL1,
            HC_ADR                  => HC_ADR,

            HCR                     => HCR,
            SOFTWARE_RESET          => SOFTWARE_RESET,
            HCFS_IN                 => HCFS_MC,
            HCFS_OUT                => HCFS_HC,
            PSSC_1                  => PSSC_1,
            PSSC_2                  => PSSC_2,
            SOC                     => SOC,
            CLKNS                   => CLKNS,
            AOCEN                   => AOCEN,

            UE                      => UE,
            RD                      => RD,
            SF                      => SF,
            SO                      => SO,
            
            FSMPS                   => FSMPS,
            FR                      => FR,
            FR_DEC                  => FR_DEC,
            LST                     => LST,
            HC_SUSPND               => HC_SUSPND,
            ATL_INT                 => ATL_INT,

            OPBERR                  => OPBERR
        );

    I_ROOTHUB: ROOTHUB_2PORT
        port map(
            CLK_48MHz               => CLK_48MHz,
            RESET                   => RESET,

            DP1_IN                  => DP1,
            DM1_IN                  => DM1,
            DP2_IN                  => DP2,
            DM2_IN                  => DM2,
            DP1_OUT                 => DP1_OUT,
            DM1_OUT                 => DM1_OUT,
            DPM1_EN                 => DPM1_EN,
            DP2_OUT                 => DP2_OUT,
            DM2_OUT                 => DM2_OUT,
            DPM2_EN                 => DPM2_EN,

            SPEED_Tx                => SPEED_Tx,
            DP_IN                   => DP_IN,
            DM_IN                   => DM_IN,
            DPM_EN                  => DPM_EN,
            DP_OUT                  => DP_OUT, 
            DM_OUT                  => DM_OUT, 

            PSW1n                   => PSW1n,
            PSW2n                   => PSW2n,
            OC1n                    => OC1n,
            OC2n                    => OC2n,

            CCS_1                   => CCS_1,
            LSDA_1                  => LSDA_1,
            PESC_1                  => PESC_1,
            POCI_1                  => POCI_1,
            PRSC_1                  => PRSC_1,
            PSSC_1                  => PSSC_1,
            PPS_1                   => PPS_1,
            PRS_1                   => PRS_1,

            CCS_2                   => CCS_2,
            LSDA_2                  => LSDA_2,
            PESC_2                  => PESC_2,
            POCI_2                  => POCI_2,
            PRSC_2                  => PRSC_2,
            PSSC_2                  => PSSC_2,
            PPS_2                   => PPS_2,
            PRS_2                   => PRS_2,

            RHSC                    => RHSC,
            UPSTREAM_P1_P2n         => UPSTREAM_P1_P2n
        );
end STRUCTURE;

-- The USB1160 is switched to little endian. The core is big endian and the switch from little to big endian is on the data input in the MC interface. This ensures
-- bit compatibility with the 1160 data sheet.
-- CLK_RDY issue fixed.
-- CommandStatus register is now 32 bit wide.
-- A software reset does not cause the CommandStatus register bit 0 to be set. The host
-- controller clears this bit if a reset is requested by setting this bit by the host controller driver.
